"""
Connector module

This module contains the Connector CLI wrapper and will act as the main interface point
for most integrators.
"""

import inspect
import json
import logging
import os
import tomllib
from enum import Enum
from pathlib import Path
from typing import Annotated, Callable

import requests
from pydantic import BaseModel, ValidationError
from rich import print, print_json
from rich.console import Console
from rich.logging import RichHandler
from typer import Exit, Option, Typer
from typer.main import get_command_name

from .models.callback import CallbackResponse
from .models.configuration import Configuration, Settings
from .models.credentials import Credential

logger = logging.getLogger("tenint.connector")


class ConfigurationError(Exception):
    pass


class LogLevel(str, Enum):
    notset = "notset"
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


class Connector:
    """
    The Connector class handles wrapping the connector code into a compliant connector
    commandline for use in integrating into the tenable integration framework.

    Args:
        settings:
            A pydantic BaseModel class describing how the connector settings.
        credentials:
            A list of Credential classes detailing the required credentials needed to
            run the connector.
    """

    app: Typer
    config: Configuration

    def __init__(
        self,
        settings: type[Settings],
        defaults: Settings | None = None,
        credentials: list[type[Credential]] | None = None,
    ):
        self.logfile = Path("job.log").absolute()
        self.console = Console(file=self.logfile.open("a", encoding="utf-8"), width=135)
        self.app = Typer(add_completion=False)
        self.settings = settings
        self.defaults = (
            defaults if defaults else settings.model_construct().model_dump()
        )
        self.credentials = credentials if credentials else []

        # Iterate through the methods on the class and construct the command-line
        # commands using the method name.  We will only want to inject commands into
        # typer that start with the cmd_ prefix.  NOTE this code has been shamelessly
        # ripped from an issue in the Typer GH repo:
        #
        # https://github.com/fastapi/typer/issues/309#issuecomment-1937836010
        #
        # e.g. cmd_config -> config
        for method, func in inspect.getmembers(self, predicate=inspect.ismethod):
            if method.startswith("cmd_"):
                name = get_command_name(method.removeprefix("cmd_"))
                self.app.command(name=name)(func)

    def job(self, func: Callable) -> Callable:
        """
        The Connector job decorator.
        """
        self.main = func
        return func

    def log_env_vars(self) -> None:
        clean = []

        # Build the hidden list
        for cred in self.credentials:
            clean += cred.env_secrets()

        for key, value in os.environ.items():
            if key in clean:
                value = "{{ HIDDEN }}"
            logger.debug(f'EnvVar {key}="{value}"')

    def fetch_config(
        self,
        data: str | None = None,
        fn: Path | None = None,
    ) -> BaseModel:
        """
        Fetch and validate the configuration from either the data string or the filepath
        and return the settings object to the caller.

        Args:
            data:
                The string object of the
        """
        settings = None

        # If a string object is passed, we will handle that first.
        if data:
            settings = self.settings(**json.loads(data))

        # If a file has been passed in instead an the suffix appears to be a JSON
        # suffix, then we will assume a JSON file and handle accordingly.
        elif fn and fn.is_file() and fn.suffix.lower() in [".json", ".jsn"]:
            with fn.open("r", encoding="utf-8") as fobj:
                settings = self.settings(**json.load(fobj))

        # If the file passed has a TOML suffix, we will process as a toml file through
        # tomllib.
        elif fn and fn.is_file() and fn.suffix.lower() in [".toml", ".tml"]:
            with fn.open("rb") as fobj:
                settings = self.settings(**tomllib.load(fobj))

        # If we processed anything, then return the settings object, otherwise raise a
        # ConfigurationError
        if settings:
            logger.debug(f"Job config={settings.model_dump(mode='json')}")
            return settings
        raise ConfigurationError("No valid configurations passed.")

    def cmd_config(
        self, pretty: Annotated[bool, Option(help="Pretty format the response")] = False
    ):
        """
        Return the connector configuration
        """
        indent = 2
        if not pretty:
            indent = None

        class Config(Configuration):
            settings_model: type[Settings] = self.settings
            credential_models: list[type[Credential]] = self.credentials

        print_json(
            Config(defaults=self.defaults).model_dump_json(
                include=["settings", "credentials", "defaults"]
            ),
            indent=indent,
        )

    def cmd_validate(self):
        """
        Validate the connector settings and credentials
        """
        print("Not yet implemented")

    def cmd_run(
        self,
        json_data: Annotated[
            str | None,
            Option(
                "--json",
                "-j",
                envvar="CONFIG_JSON",
                help="The JSON config object as a string",
            ),
        ] = None,
        filename: Annotated[
            Path | None,
            Option(
                "--filename",
                "-f",
                envvar="CONFIG_FILENAME",
                help="Filename of either a json or toml file containing the job config",
            ),
        ] = None,
        job_id: Annotated[
            str | None,
            Option(
                "--job-id",
                "-J",
                envvar="JOB_ID",
                help="The unique job id of this invocation",
            ),
        ] = None,
        callback_url: Annotated[
            str | None,
            Option(
                "--callback",
                "-c",
                envvar="CALLBACK_URL",
                help="The URL the connector should call back to when it completes",
            ),
        ] = None,
        log_level: Annotated[
            LogLevel,
            Option(
                "--log-level",
                "-l",
                envvar="LOG_LEVEL",
                help="Sets the logging verbosity for the job",
                case_sensitive=False,
            ),
        ] = LogLevel.info,
        since: Annotated[
            int | None,
            Option(
                "--since",
                "-s",
                envvar="SINCE",
                help="When did the last successful run start?",
            ),
        ] = None,
    ):
        """
        Invoke the connector
        """
        logging.basicConfig(
            level=log_level.upper(),
            # format="%(asctime) %(name)s(%(filename)s:%(lineno)d): %(message)s",
            format="%(name)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            handlers=[
                # logging.FileHandler("job.log"),
                RichHandler(
                    console=self.console,
                    show_path=False,
                    rich_tracebacks=True,
                    tracebacks_show_locals=True,
                    omit_repeated_times=False,
                ),
                logging.StreamHandler(),
            ],
        )
        self.log_env_vars()
        logger.info(f"Logging to {self.logfile}")
        status_code = 0
        resp = None

        # Attempt to run the connector job and handle any errors that may be thrown
        # in a graceful way.
        try:
            config = self.fetch_config(json_data, filename)
            resp = self.main(config=config, since=since)
        except (ValidationError, ConfigurationError) as e:
            logging.error(f"Invalid configuration presented: {e}")
            status_code = 2
        except Exception as _:
            logging.exception("Job run failed with error", stack_info=2)
            status_code = 1

        # Set the Callback payload to the job response if the response is a dictionary
        try:
            payload = CallbackResponse(exit_code=status_code, **resp).model_dump(
                mode="json"
            )
        except (ValidationError, TypeError) as _:
            logger.error(f"Job response format is invalid: {resp=}")
            payload = CallbackResponse(exit_code=status_code).model_dump(mode="json")

        # If a callback and a job id have been set, then we will initiate a callback
        # to the job manager with the response payload of the job to inform the manager
        # that we have finished.
        if job_id and callback_url:
            requests.post(
                callback_url,
                headers={"X-Job-ID": job_id},
                json=payload,
                verify=False,
            )
            logger.info(f"Called back to {callback_url=} with {job_id=} and {payload=}")
        else:
            logger.warning("Did not initiate a callback!")
        logger.info(f"callback={payload}")

        # Exit the connector with the status code from the runner.
        raise Exit(code=status_code)



import connector
from typer.testing import CliRunner

runner = CliRunner()


def test_connector_run():
    result = runner.invoke(
        connector.connector.app, ["run", "-l", "INFO", "-j", '{"is_bool": true}']
    )
    print(result.stdout)
    assert result.exit_code == 0
    assert "hello world" in result.stdout



from pydantic import BaseModel


class CallbackDataTypeCounts(BaseModel):
    sent: int = 0
    modified: int = 0
    failed: int = 0


class CallbackCounters(BaseModel):
    assets: CallbackDataTypeCounts = CallbackDataTypeCounts()
    findings: CallbackDataTypeCounts = CallbackDataTypeCounts()


class CallbackResponse(BaseModel):
    exit_code: int
    counts: CallbackCounters = CallbackCounters()



from typing import Annotated, Any, Literal

from pydantic import AnyHttpUrl, BaseModel, ConfigDict, Field, SecretStr, computed_field


def credential_schema(schema: dict[str, Any]) -> None:
    """
    Credential model schema adjustment.

    A simple function to remove the attributes that shouldn't be part of the UI schema
    from the schema itself.
    """
    for attr in ("prefix", "name", "slug", "definition", "description"):
        schema.get("properties", {}).pop(attr, None)


class Credential(BaseModel):
    """
    Credential base class
    """

    model_config = ConfigDict(json_schema_extra=credential_schema)
    description: str
    prefix: str
    name: str
    slug: str

    @computed_field
    @property
    def definition(self) -> dict[str, Any]:
        return self.model_json_schema()

    @classmethod
    def env_secrets(cls) -> list[str]:
        resp = []
        schema = cls.model_json_schema()
        prefix = cls.model_fields["prefix"].default
        for name, props in schema["properties"].items():
            if props.get("format") == "password":
                resp.append(f"{prefix}_{name}".upper())
        return resp


class TenableCloudCredential(Credential):
    """
    Tenable Cloud Credential
    """

    prefix: Literal["tio"] = "tio"
    name: Literal["Tenable Cloud"] = "Tenable Cloud"
    slug: Literal["tvm"] = "tvm"
    description: str = "Tenable Cloud Credential"
    url: Annotated[
        AnyHttpUrl,
        Field(
            title="Cloud URL",
            description="Only needs to be modified when interacting with a FedRAMP cloud instance",
        ),
    ] = "https://cloud.tenable.com"
    access_key: Annotated[
        str,
        Field(title="API Access Key", description="Tenable Cloud API Access Key"),
    ]
    secret_key: Annotated[
        SecretStr,
        Field(title="API Secret Key", description="Tenable Cloud API Secret Key"),
    ]


class TenableSCCredential(Credential):
    """
    Tenable Security Center Credential
    """

    prefix: Literal["tsc"] = "tsc"
    name: Literal["Tenable Security Center"] = "Tenable Security Center"
    slug: Literal["tsc"] = "tsc"
    description: str = "Tenable Security Center Credential"
    url: Annotated[
        AnyHttpUrl,
        Field(title="Host URL", description="Security Center Application URL"),
    ]
    access_key: Annotated[
        str,
        Field(title="API Access Key", description="Security Center API Access Key"),
    ]
    secret_key: Annotated[
        SecretStr,
        Field(title="API Secret Key", description="Security Center API Secret Key"),
    ]



import logging
from typing import Annotated

from pydantic import AnyHttpUrl, Field

from tenint import Connector, Credential, Settings


class TestCredential(Credential):
    prefix: str = "test"
    name: str = "Test Credential"
    slug: str = "test"
    api_key: Annotated[
        str,
        Field(title="API Key", description="Application API Key"),
    ]
    url: Annotated[
        AnyHttpUrl,
        Field(title="Application URL", description="Application URL"),
    ] = "https://httpbin.org/get"


class AppSettings(Settings):
    is_bool: Annotated[
        bool,
        Field(title="Yes?", description="An example boolean flag"),
    ] = True


log = logging.getLogger("test-connector")
connector = Connector(settings=AppSettings, credentials=[TestCredential])


@connector.job
def main(config: AppSettings, since: int | None = None) -> dict:
    log.debug("This is a debug test")
    log.info("this is an info test")
    log.warning("this is a warning")
    log.error("this is an error")
    print("hello world")
    return {"counts": {"assets": {"sent": 0}}}


if __name__ == "__main__":
    connector.app()



from typing import Any

from pydantic import BaseModel, ConfigDict, Field, computed_field

from .credentials import Credential


class Settings(BaseModel):
    model_config = ConfigDict(extra="forbid")


class Configuration(BaseModel):
    settings_model: type[Settings] = Field(exclude=True)
    credential_models: list[type[Credential]] = Field(exclude=True, default=[])
    defaults: dict

    @computed_field
    @property
    def settings(self) -> dict[str, Any]:
        return self.settings_model.model_json_schema()

    @computed_field
    @property
    def credentials(self) -> list[dict[str, Any]]:
        return [
            C.model_construct().model_dump(
                mode="json",
                include=["name", "slug", "prefix", "definition"],
            )
            for C in self.credential_models
        ]



import os
import subprocess
import tomllib
from pathlib import Path
from typing import Annotated

from rich.console import Console
from rich.panel import Panel
from typer import Option, Typer

from tenint.models.marketplace import MarketplaceConnector
from tenint.models.pyproject import PyProject

console = Console()
tmpl = Path(os.path.dirname(__file__)).joinpath("templates")
app = Typer()


@app.command("init")
def init_connector(
    path: Annotated[
        Path, Option(help="initialization path", dir_okay=True, file_okay=False)
    ] = Path("."),
):
    is_dirty = False
    for fn in ("pyproject.toml", "connector.py"):
        src = tmpl.joinpath(fn)
        dest = path.joinpath(fn)

        if dest.exists():
            console.print(
                f"[yellow bold]WARNING[/yellow bold]: skipped "
                f"existing [cyan]{dest}[/cyan] as it already exists"
            )
            is_dirty = True
        else:
            with src.open("rb") as sobj, dest.open("wb") as dobj:
                dobj.write(sobj.read())

    tests = path.joinpath("tests")
    if not tests.exists():
        tests.mkdir()
        src = tmpl.joinpath("test_connector.py")
        dest = tests.joinpath("test_connector.py")
        with dest.open("wb") as dobj, src.open("rb") as sobj:
            dobj.write(sobj.read())
    else:
        console.print(
            "[yellow bold]NOTE[/yellow bold]: skipped adding tests.  tests folder exists."
        )

    console.print(
        Panel(
            "Now that you have an initialized tenint connector, please review the "
            "files that have been created for your next steps.\n\n"
            "- [bold cyan]pyproject.toml[/bold cyan] contains the project "
            " requirements and is used to build the marketplace files\n"
            "- [bold cyan]connector.py[/bold cyan] is a sample connector "
            "to get you started.  It contains some example settings, a "
            "credential example, and a sample function.",
            title="[bold cyan]Tenable Integration Framework Connector[/bold cyan]",
        )
    )
    if is_dirty:
        console.print(
            Panel(
                "Could not initialize all of the files as some appeared to already "
                "exist.  Any files that previously existed were not modified.  "
                "Recommend manually reviewing your project to ensure everything "
                "looks correct.",
                title="[bold red]Dirty Environment Warning[/bold red]",
                style="yellow",
            )
        )


@app.command("build")
def build_connector(
    path: Annotated[Path, Option(help="connector code path")] = Path("."),
    platform: Annotated[str | None, Option(help="platform to build to")] = None,
    tag: Annotated[
        str | None,
        Option(help="Tag for the fimal image. defaults to the project name."),
    ] = None,
    cleanup: Annotated[
        bool | None, Option(help="auto-remove any generated build files?")
    ] = None,
):
    tmpldfile = tmpl.joinpath("Dockerfile")
    dockerfile = path.joinpath("Dockerfile")
    pyproject = path.joinpath("pyproject.toml")

    if not tag:
        with pyproject.open("rb") as pf:
            project = PyProject(**tomllib.load(pf))
            tag = project.project.name

    if not dockerfile.exists():
        with tmpldfile.open("rb") as src, dockerfile.open("wb") as dst:
            dst.write(src.read())
            cleanup = True if cleanup is None else cleanup

    builder = subprocess.Popen(
        ["docker", "build", "-t", tag, path],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    for line in builder.stdout:
        console.out(line.decode("utf-8").strip("\n"))

    if cleanup:
        dockerfile.unlink()


@app.command("marketplace")
def gen_marketplace(
    image: Annotated[str, Option(..., help="Docker Image Name")] = None,
    icon: Annotated[
        str, Option(..., help="Icon Image URL")
    ] = "https://nourl.example/logo.svg",
    path: Annotated[Path, Option(help="connector code path")] = Path("."),
    output: Annotated[Path, Option(help="output marketplace json file")] = None,
):
    pyproject_file = path.joinpath("pyproject.toml")
    with pyproject_file.open("rb") as fobj:
        project = PyProject(**tomllib.load(fobj))

    if not image:
        image = project.project.name

    mpfile = path.joinpath("marketplace-object.json")
    mp = MarketplaceConnector.load_from_pyproject(
        filename=project, image_url=image, icon_url=icon
    ).model_dump(mode="json")
    console.print(mp)
    if output:
        with mpfile.open("w", encoding="utf-8") as fobj:
            fobj.write(mp.model_dump_json())



from __future__ import annotations

import tomllib
from pathlib import Path

from pydantic import AnyHttpUrl, BaseModel, EmailStr

from .pyproject import PyProject


class MarketplaceConnector(BaseModel):
    name: str
    slug: str
    description: str
    icon_url: AnyHttpUrl
    image_url: str
    timeout: int
    marketplace_tag: str
    connector_owner: str
    support_contact: EmailStr
    tags: list[str]

    @classmethod
    def load_from_pyproject(
        cls,
        filename: PyProject | Path | str,
        image_url: str,
        icon_url: str,
    ) -> MarketplaceConnector:
        """
        Builds the marketplace object off of the PyProject file.

        Args:
            filename:
                The filename of the pyproject.toml file.

        Returns:
            Marketplace:
                The marketplace object.

        Example:

            >>> marketplace = Marketplace.load_from_pyproject('pyproject.toml')
        """
        if isinstance(filename, PyProject):
            obj = filename
        else:
            if isinstance(filename, str):
                filename = Path(filename)
            with filename.open("rb") as fobj:
                obj = PyProject(**tomllib.load(fobj))
        return cls(
            name=obj.tool.tenint.connector.title,
            slug=obj.project.name,
            description=obj.project.description,
            icon_url=icon_url,
            image_url=image_url,
            timeout=obj.tool.tenint.connector.timeout,
            marketplace_tag=obj.project.version,
            connector_owner=obj.project.authors[0].name,
            support_contact=obj.project.authors[0].email,
            tags=obj.tool.tenint.connector.tags,
        )



from .models.credentials import Credential, TenableCloudCredential, TenableSCCredential  # noqa F401
from .models.configuration import Configuration, Settings  # noqa F401
from .connector import Connector  # noqa F401

__version__ = "1.0.3"



from .configuration import Configuration, Settings  # noqa F401
from .credentials import Credential, TenableCloudCredential, TenableSCCredential  # noqa F401
from .pyproject import PyProject  # noqa F401



from typing import Annotated

from pydantic import BaseModel, Field
from pydantic.networks import AnyHttpUrl, EmailStr


class TestingDependencies(BaseModel):
    testing: list[str]


class ProjectAuthor(BaseModel):
    name: str
    email: EmailStr


class TenintConnectorUrls(BaseModel):
    support: AnyHttpUrl


class Project(BaseModel):
    name: str
    description: str
    version: str
    authors: list[ProjectAuthor]
    urls: TenintConnectorUrls
    dependencies: list[str]
    optional_dependencies: Annotated[
        TestingDependencies, Field(alias="optional-dependencies")
    ]


class TenintConnector(BaseModel):
    title: str
    tags: list[str]
    timeout: int = 3600


class Tenint(BaseModel):
    connector: TenintConnector


class Tools(BaseModel):
    tenint: Tenint


class PyProject(BaseModel):
    project: Project
    tool: Tools
